from __future__ import annotations as _annotations

from upsonic.profiles import ModelProfile


def grok_model_profile(model_name: str) -> ModelProfile | None:
    """Get the model profile for a Grok model."""
    return None